# your cluster address, if local probably 'localhost:9200'
ES_HOST = 'https://beba020ee88d49488d8f30c163472151.eu-west-2.aws.cloud.es.io:9243/'
ES_USE_AUTH = True
ES_USER = 'admin_pandagg'
ES_PASSWORD = 'X(2L*4.z]c_Hjw@'

# where your table dumps are stored, and where serialized output will be written
DATA_DIR = '/Users/leonard/Alkemics/studies/initiatives/pandagg/data/imdb'
OUTPUT_FILE_NAME = 'serialized.json'
